/*
 * pka_functions.h
 *
 * API to use the large number arithmetic functions
 * supported by PKA engine.
 */

#ifndef __PKA_FUNCTIONS_H__
#define __PKA_FUNCTIONS_H__

#include <sys/platform.h>
#include <stdbool.h>

#ifndef __DEFINED_MAX
static int max(int __a, int __b){
	if(__a >= __b)
		return __a;
	else
		return __b;
}
#endif

/*
 * pka_mult():  large number multiplication.  Blocking function.
 *
 *	input:
 *		vectA   - pointer to buffer storing multiplicand
 *		vectAsz - size of multiplicand in 32-bit words
 *		vectB   - pointer to buffer storing multiplier
 *		vectBsz - size of multiplier in 32-bit words
 *		result  - pointer to buffer holding result (size of buffer should be at least vectAsz + vectBsz)
 *	output:
 *	    function returns the pointer to result buffer if successful and NULL on error
 */
uint32_t *pka_mult(uint32_t *vectA, uint32_t vectAsz, uint32_t *vectB, uint32_t vectBsz, uint32_t *result);


/*
 * pka_div():  large number division.  Blocking function.
 *
 *	input:
 *		vectA   - pointer to buffer storing dividend
 *		vectAsz - size of multiplicand in 32-bit words
 *		vectB   - pointer to buffer storing divisor
 *		vectBsz - size of multiplier in 32-bit words
 *		result  - pointer to buffer holding result (if NULL, result will not be copied)
 *		remainder - pointer to the buffer holding remainder (if NULL, result will not be copied)
 *	output:
 *	    function returns the pointer to result buffer if successful and NULL on error
 */
uint32_t *pka_div(uint32_t *vectA, uint32_t vectAsz, uint32_t *vectB, uint32_t vectBsz, uint32_t *result, uint32_t *remainder);

/*
 * pka_add():  large number addition.  Blocking function.
 *
 *	input:
 *		vectA   - pointer to buffer storing addend
 *		vectAsz - size of multiplicand in 32-bit words
 *		vectB   - pointer to buffer storing addend
 *		vectBsz - size of multiplier in 32-bit words
 *		result  - pointer to buffer holding sum (size of buffer should be at least MAX(vectAsz, vectBsz)+1 )
 *	output:
 *	    function returns the pointer to result buffer if successful and NULL on error
 */
uint32_t *pka_add(uint32_t *vectA, uint32_t vectAsz, uint32_t *vectB, uint32_t vectBsz, uint32_t *result);


/*
 * pka_sub():  large number subtraction.  Blocking function.
 *
 *	input:
 *		vectA   - pointer to buffer storing minuend
 *		vectAsz - size of multiplicand in 32-bit words
 *		vectB   - pointer to buffer storing subtrahend
 *		vectBsz - size of multiplier in 32-bit words
 *		result  - pointer to buffer holding difference
 *
 *		requirements: vectA >= VectB
 *
 *	output:
 *	    function returns the pointer to result buffer if successful and NULL on error
 */
uint32_t *pka_sub(uint32_t *vectA, uint32_t vectAsz, uint32_t *vectB, uint32_t vectBsz, uint32_t *result);



/*
 * pka_cmp():  large number comparison.  Blocking function.
 *
 *	input:
 *		vectA   - pointer to buffer storing minuend
 *		vectB   - pointer to buffer storing subtrahend
 *		vectSz - size of numbers in 32-bit words
 *	output:
 *	    0 if A = B
 *	    1 if A > B
 *	   -1 if A < B
 *	   -2 on error
 */
int32_t pka_cmp(uint32_t *vectA, uint32_t *vectB, uint32_t vectSz);

/*
 *
 * pka_modinv(): Compute modular inverse of large number.  Blocking function.
 * A ^-1 mod B -> D
 *
 *	input:
 *		vectA   - pointer to buffer storing number to find inverse of
 *		vectAsz - size of number in 32-bit words
 *		vectB   - pointer to buffer storing modulus
 *		vectBsz - size of modulus in 32-bit words
 *	    result  - pointer to buffer storing result
 *
 * Input restrictions:
 * 1. 0 < A_Len <= Max_Len
 * 2. 0 < B_Len <= Max_Len
 * 3. Modulus in B must be ODD
 * 4. Modulus in B must not have a value of '1' (result is undefined with not error indicated)
 * 5. The highest word of the modulus vector, as indicated by B_Len, may not be 0
 *
 * Overlap:
 * Scratchpad area starting at D may not overlap with any of the other vectors
 *
 * Also, during execution of a ModInv operation, the last 34 bytes of the PKA RAM are used as general
 * scratchpad for the Sequencers program execution. This (fixed location) area may not overlap with any of the
 * input vectors and/or the D vector scratchpad area during execution.
 *
 * Return code in PKA_SHIFT reg:
 *  0 -> success; vector D hold result
 *  7 -> no inverse exists (GCD(A,B) != 1, i.e. A and B have common factors); result undefined
 *  31 -> error.  Modulus EVEN; result undefined
 *
 */
uint32_t pka_modinv(uint32_t *vectA, uint32_t vectAsz, uint32_t *vectB, uint32_t vectBsz, uint32_t *result);


/************
 * pka_modexp(): Modular exponentiation.  Blocking function.
 * C^A mod B -> D
 *
 *	input:
 *		exp  (A)- pointer to exponent
 *		mod  (B)- pointer to modulus
 *		base (C)- pointer to base
 *		szWds - size of vectors in 32-bit words.  Exponent, modulus and base must all be same size.
 *	output:
 *		result (D) - pointer to output buffer
 *
 * Input restrictions:
 * 	1. 0 < A_Len <= Max_Len
 * 	2. 1 < B_len <= Max_Len
 * 	3. Modulus B must be odd
 * 	4. Modulus B > 2^32
 * 	5. Base C < Modulus B
 * 	6. Vectors B and C must be followed by an empty 32-bit 'buffer' word
 *
 * **************/

uint32_t *pka_modexp(uint32_t *exp, uint32_t expSzWds, uint32_t *mod, uint32_t *base, uint32_t ModBaseSzWds, uint32_t *result);

/************
 * pka_modexpcrt(): Modular exponentiation using Chinese Remainder Theorem.  Blocking function.
 * C^A mod B -> D
 *
 *	input:
 *		dp   (A)- pointer to dp (= d mod(p-1))
 *		dq   (A)- pointer to dq (= d mod(q-1))
 *		p    (B)- pointer to p
 *		q    (B)- pointer to q
 *		qinv (C)- pointer to q inverse
 *		base (D)- pointer to input
 *		szWds - size of vectors in 32-bit words.  Exponent, modulus and base must all be same size.
 *	output:
 *		result (D) - pointer to output buffer
 *		           - computes:
 *		                     m1 = base^dp mod p
 *		                     m2 = base^dq mod q
 *		                     h  = ((m1 - m2) * qinv) mod p
 *		                     result = m2 + q * h
 *
 *
 * Input restrictions:
 * 	1. 0 < A_Len <= Max_Len
 * 	2. 1 < B_len <= Max_Len
 * 	3. Mod P and Mod Q must be odd
 * 	4. Mod P > Mod Q > 2^32
 * 	5. Mod P and Mod Q must be co-prime (GCD=1)
 * 	6. 0 < dp < (Mod P-1)
 * 	7. 0 < dq < (Mod Q-1)
 * 	8. qinv * q = 1 MOD P
 * 	9. base < (Mod P * Mod Q)  (i.e. < n)
 * 	10. Mod P and Mod Q must be followed by an empty "buffer" word
 *
 * **************/
uint32_t *pka_modexpcrt(uint32_t *dp, uint32_t *dq, uint32_t *p, uint32_t *q, uint32_t *qinv, uint32_t *ciphertxt, uint32_t szWds, uint32_t *plaintxt) ;

#endif /* __PKA_FUNCTIONS_H__ */
